import type { NamespaceTable } from "./NameSpaceTable";
import { coerceNodeId, NodeId } from "./NodeId";
import { UAReference } from "./UAReference";
import { assert } from "@/util/assert";
import { XMLElem, type IToXML } from "@/util/XmlElem";
import { UARolePermission } from "./UARolePermission";
import { UAExtension } from "./UAExtension";
import { UAUserWriteMask } from "./UAUserWriteMask";
import { UAWriteMask } from "./UAWriteMask";
import { UAAccessRestriction } from "./UAAccessRestriction";
import { type IAddressSpace } from "./IAddressSpace";
import { ReferenceTypeIds } from "./opcua_node_ids";
export class UABaseNode implements IToXML{

    public static nullBaseNode=new UABaseNode({browseName: "", addressSpace: {} as IAddressSpace, nodeId: NodeId.nullNodeId});


    public nodeId: NodeId;
    public nodeClass="UABaseNode";
    public browseName: string;
    public addressSpace: IAddressSpace
    public displayName?: string; //LocText
    public description?: string; //LocText
    public symbolicName?: string; //SymbolicName
    public releaseStatus?: string; //ReleaseStatus
    public hasNoPermissions?: boolean; 
    public writeMask?: UAWriteMask; 
    public userWriteMask?: UAUserWriteMask;
    public category?: string;
    public documentation?: string;
    public accessRestriction?: UAAccessRestriction; 
    public rolePermissions: UARolePermission[]=[]; 
    public extensions: UAExtension[]=[];
    public references: UAReference[]=[];
    public opts: string[]=[]; 
    
    constructor(options: UABaseNodeOptions) {
        this.nodeId=options.nodeId;
        this.browseName=options.browseName;
        this.addressSpace=options.addressSpace;
        this.displayName=options.displayName;
        this.references=options.references||[];     
    }


    reIndex(nst: NamespaceTable, onst: NamespaceTable) {
        const nsName=onst.getUri(this.nodeId.namespace);
        assert(nsName!=undefined)
        const newIndex=nst.getIndex(nsName);
        assert(newIndex!=undefined)
        this.nodeId.namespace=newIndex;

        for(const uaref of this.references) {
            uaref.reIndex(nst, onst);
        }
    }

    getParent(): UABaseNode|null {
        const ref=this.getParentRef();
        if(ref?.fromNode===this)
            return ref.toNode;
        if(ref?.toNode===this)
            return ref.fromNode;
        return null;
    }

    setParent(newParentNode: UABaseNode, newRefType: string) {
        const hierReferences=this.addressSpace.getSubTreeAsList("ns=0;i="+ReferenceTypeIds.HierarchicalReferences);
        let href;
        for(const r of this.references) { //find current href in references
            href=hierReferences.find((n) => {n.browseName===r.referenceType})
        }
        for(let i=this.references.length;i--;i>=0) { //remove current href from references
            const r=this.references[i];
            if(r.referenceType==href?.browseName)
                this.references.splice(i,1);
        }
        const newRefA=new UAReference(this.nodeId, newRefType, newParentNode.nodeId, true);
        newRefA.fromNode=this;
        newRefA.toNode=newParentNode;
        this.references.push(newRefA);
        const newRefB=new UAReference(newParentNode.nodeId, newRefType, this.nodeId, false); 
        newRefB.fromNode=this;
        newRefB.toNode=newParentNode;
        this.references.push(newRefB);
    }

    getParentRef(): UAReference|null{
        for(const ref of this.references) {
            switch(ref.referenceType) {
                case 'HasComponent':
                case 'HasOrderedComponent':
                case 'Organizes':
                case 'HasProperty':
                case 'HasSubtype':
                case 'HasAddIn':
                    if(ref.fromNode===this&&ref.isForward===false)
                        return ref;
                    if(ref.toNode===this&&ref.isForward===true)
                        return ref;
            }
        }
        return null;
    }

    getChildren(): UABaseNode[] {
        const children: UABaseNode[]=[];
        for(const ref of this.references) {
            switch(ref.referenceType) {
                case 'HasComponent':
                case 'HasOrderedComponent':
                case 'Organizes':
                case 'HasProperty':
                case 'HasSubtype':
                case 'HasAddIn':
                    if(ref.isForward&&ref.fromNode===this) {
                        if(!children.includes(ref.toNode))
                            children.push(ref.toNode);
                    } 
                    if(!ref.isForward&&ref.toNode===this) {
                        if(!children.includes(ref.fromNode))
                            children.push(ref.fromNode);
                    }
                    break;
                }
        }
        return children;
    }

    getModellingRule(): string|undefined{
        this.references.forEach((ref)=>{
            if(ref.referenceType == "HasModellingRule"){
                return ref.toNode.displayName;
            }
        });
        return undefined;
    }

    resolveReferences(nm: Map<string, UABaseNode>) {
        for(const ref of this.references) {
            const fromNode=nm.get(ref.fromRef.toString()) 
            if(fromNode)
                ref.fromNode=fromNode;
            const toNode=nm.get(ref.toRef.toString()) 
            if(!toNode) 
                continue; //TODO: if we cant find the node; the parser is still incomplete or the nodeset is broken
            
            ref.toNode=toNode;
            if(ref.fromNode.nodeId.toString()===this.nodeId.toString()){ //add this reference to referenced node
                //Bug? when loading from filedrop; fromNode is a proxy; this is not.
                if(!ref.toNode.references.includes(ref))
                    ref.toNode.references.push(ref);
            } 
        }
    }



    static fromXML(xmlObject: any, addressSpace: IAddressSpace): UABaseNode{
        const xmlReferences=xmlObject['References']||[];
        const references:UAReference[]=[];
        const nodeId=coerceNodeId(xmlObject['@_NodeId']);
        for(const xmlref of xmlReferences.Reference||[]) {
            references.push(UAReference.fromXML(xmlref, nodeId));
        }
        const ua=new UABaseNode({nodeId: nodeId,
                                browseName: xmlObject['@_BrowseName'], 
                                displayName: xmlObject['DisplayName']['#text'], 
                                references: references,
                                addressSpace: addressSpace
                            }); 


        return ua;
    }

    toXML(_lnst: NamespaceTable, _gnst: NamespaceTable): XMLElem {
        throw new Error("UABaseNode has no xml rep; implement in subtype.");
    }

    static localNodeId(nodeId: NodeId, lnst: NamespaceTable, gnst: NamespaceTable) {
        const ns_uri=gnst.getUri(nodeId.namespace);
        assert(ns_uri);
        const nsIdx=lnst.getIndex(ns_uri);
        assert(nsIdx);
        const tmpNode= coerceNodeId(nodeId.toString());
        tmpNode.namespace=nsIdx;
        return tmpNode;
    } 
}


export interface UABaseNodeOptions {
    browseName: string;
    nodeId: NodeId;
    namespace?: string;
    references?: UAReference[];
    displayName?: string;
    description?: string;
    addressSpace: IAddressSpace;
}